////////////////////////////////////////////////////////////////
// Ce module contient le protocole d'echange entre un client
// et un serveur. Il s'appuie sur le module socket.cpp
// pour gerer l'envoi/reception de donnees avec des sockets
// non bloquantes
//
// Sylvain MARECHAL - sylvain.marechal1@libertysurf.fr
////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "pongcore.h"
#include "pongsock.h"

// #define 
#define SEND_DIMENSIONS		'D'
#define SEND_RIGHTBAT_POS	'R'
#define SEND_LEFTBAT_POS	'L'
#define SEND_BALL_POS		'B'
#define SEND_SERVEUR_STATUS	'S'

// Variables globales
// Gestion des sockets
static SOCKET s_hSocket = INVALID_SOCKET;
static char * s_pBufferSend = NULL;
static int s_cbBufferSend = 0;
static int s_szBufferRecv[256];
static int s_cbBufferRecv = 0;

// Gestion de l'echelle
static int s_xLargeurOther = 0;
static int s_yHauteurOther = 0;
static int s_yBatLenOther = 0;


////////////////////////////////////////////////////////////////
// Reinitialise les statics de ce module a 0
// memorise la valeur de la socket dans une globale static
////////////////////////////////////////////////////////////////
void PongSocketInitializeSocket( SOCKET hSocket )
{
    if( s_pBufferSend ) free( s_pBufferSend );
    s_pBufferSend = NULL;
    s_cbBufferSend = 0;
    s_cbBufferRecv = 0;
    s_hSocket = hSocket;
    s_xLargeurOther = 0;
    s_yHauteurOther = 0;
}

////////////////////////////////////////////////////////////////
// RAZ des ressources utilisees
////////////////////////////////////////////////////////////////
void PongSocketDestroySocket()
{
    PongSocketInitializeSocket( INVALID_SOCKET );
}

////////////////////////////////////////////////////////////////
// Envoi d'un buffer de donnees avec bufferisation si jamais
// on ne peut pas envoyer la donnee
////////////////////////////////////////////////////////////////
static int PongSocketSendBuffer( char * pBuffer, int cbBuffer )
{
    int cbSent = 0;
    
    
    // On commence deja par envoye ce qui ne l'a pas ete
    if( s_cbBufferSend > 0 )
    {
	cbSent = SocketSend( s_hSocket, s_pBufferSend, s_cbBufferSend );
	if( cbSent == s_cbBufferSend )
	{
	    // On a reussi a envoyer les donnees pendantes
	    s_cbBufferSend = 0;
	    free( s_pBufferSend );

	    // On ne sort pas : On doit encore envoyer
	    // ce qui nous a ete passe en argument
	}
	else if( cbSent < 0 )
	{
	    // Y'a un probleme
	    return -1;
	}
	else // cbSent < s_cbBufferSend
	{
	    int cbLeft = s_cbBufferSend - cbSent;
	    char * pSave = s_pBufferSend;
	    
	    // On ne peut pas tout envoyer : On realloue le buffer
	    s_pBufferSend = (char *)realloc( s_pBufferSend, cbLeft + cbBuffer );
	    if( s_pBufferSend == NULL )
	    {
		// PB memoire
		s_pBufferSend = pSave;
		return -1;
	    }
	    
	    // On deplace ce qui n'a pas ete envoye (cbLeft) au debut
	    memmove( s_pBufferSend, s_pBufferSend + cbSent, cbLeft );

	    // On copie ce qu'on nous a passe en argument a la fin
	    if( pBuffer ) // pBuffer peut etre NULL
	    {
		memcpy( s_pBufferSend + cbLeft, pBuffer, cbBuffer );
	    }

	    // On met a jour le compteur
	    s_cbBufferSend = cbLeft + cbBuffer;

	    // On renvoie 0, signe que y'a de la donnee pendante 
	    // a envoyer
	    return 0;
	}
    }

    // On s'occupe maintenant du buffer passe en argument
    // pBuffer peut etre NULL
    if( cbBuffer > 0 )
    {
	cbSent = SocketSend( s_hSocket, pBuffer, cbBuffer );
	if( cbSent < 0 )
	{
	    // Y'a probleme
	    return -1;
	}
	else if( cbSent < cbBuffer )
	{
	    int cbLeft = cbBuffer - cbSent;
	    
	    // On ne peut pas tout envoyer : On alloue le buffer
	    s_pBufferSend = (char *)malloc( cbLeft );
	    if( s_pBufferSend == NULL )
	    {
		// PB memoire
		return -1;
	    }
	    
	    // On copie ce qui n'a pas ete envoye (cbLeft)
	    memcpy( s_pBufferSend, pBuffer + cbSent, cbLeft );

	    // On met a jour le compteur
	    s_cbBufferSend += cbLeft;

	    // On renvoie 0, signe que y'a de la donnee pendante 
	    // a envoyer
	    return 0;
	}
	else // => cbSent == cbBuffer
	{
	}
    }

    // Tout a ete envoye
    return 1;
}

////////////////////////////////////////////////////////////////
// Envoi de la taille de l'ecran a l'autre joueur
////////////////////////////////////////////////////////////////
int PongSocketSendDims()
{
    char szBuffer[256]; int cbBuffer;

    cbBuffer = sprintf( szBuffer, "%c x=%d y=%d yBatLen=%d\n", SEND_DIMENSIONS, PongGetLargeur(), PongGetHauteur(), PongGetBatLen() );
    return PongSocketSendBuffer( szBuffer, cbBuffer );
}

////////////////////////////////////////////////////////////////
// Envoi de la position de la raquette droite (client->serveur)
////////////////////////////////////////////////////////////////
int PongSocketSendRightBatPos()
{
    char szBuffer[256]; int cbBuffer;

    cbBuffer = sprintf( szBuffer, "%c y=%d\n", SEND_RIGHTBAT_POS, PongGetRightBatPos() );
    return PongSocketSendBuffer( szBuffer, cbBuffer );
}

////////////////////////////////////////////////////////////////
// Envoi de la position de la balle
////////////////////////////////////////////////////////////////
int PongSocketSendBallPos()
{
    char szBuffer[256]; int cbBuffer;;

    cbBuffer = sprintf( szBuffer, "%c x=%d y=%d\n", SEND_BALL_POS, PongGetXBallPos(), PongGetYBallPos() );
    return PongSocketSendBuffer( szBuffer, cbBuffer );
}

////////////////////////////////////////////////////////////////
// Envoi du status du jeu (gagne/perdu)
////////////////////////////////////////////////////////////////
int PongSocketSendStatusGame( GameStatus SenderStatus )
{
    char szBuffer[256]; int cbBuffer;

    cbBuffer = sprintf( szBuffer, "%c %d\n", SEND_SERVEUR_STATUS, SenderStatus );
    return PongSocketSendBuffer( szBuffer, cbBuffer );
}

////////////////////////////////////////////////////////////////
// Envoi de la position de la raquette gauche (serveur->client)
////////////////////////////////////////////////////////////////
int PongSocketSendLeftBatPos()
{
    char szBuffer[256]; int cbBuffer;

    cbBuffer = sprintf( szBuffer, "%c y=%d\n", SEND_LEFTBAT_POS, PongGetLeftBatPos() );
    return PongSocketSendBuffer( szBuffer, cbBuffer );
}

////////////////////////////////////////////////////////////////
// Envoi des donnees bufferisees.
// A appele en reponse a l'evenement FD_WRITE
////////////////////////////////////////////////////////////////
int PongSocketEvtReSend()
{
    return PongSocketSendBuffer( NULL, 0 );
}

////////////////////////////////////////////////////////////////
// Reception d'une ligne : Decodage de l'information
// et mise a jour des nouveaux parametres du jeu
////////////////////////////////////////////////////////////////
static int PongSocketEvtRecvLine( char * pszBuffer, GameStatus *pStatus )
{
    char type; int x, y, yBatLen;
    int ret = 0;
    
    *pStatus = ENCOURS;
    switch( pszBuffer[0] )
    {
    case SEND_DIMENSIONS:
	{
	    if( sscanf( pszBuffer, "%c x=%d y=%d yBatLen=%d", &type, &x, &y, &yBatLen ) == 4 )
	    {
		// Utile pour positionner correctement
		s_xLargeurOther = x;
		s_yHauteurOther = y;
		s_yBatLenOther = yBatLen;

	    }
	    break;
	}
    case SEND_RIGHTBAT_POS:
	{
	    if( sscanf( pszBuffer, "%c y=%d", &type, &y ) == 2 )
	    {
		// On met la bonne echelle
		y = y + s_yBatLenOther/2;
		if( s_yHauteurOther ) y = (y * PongGetHauteur()) / s_yHauteurOther;
		y = y - PongGetBatLen() / 2;
		
		// Positionnement de la raquette droite
		PongSetRightBatPos( y );
	    }
	    break;
	}
    case SEND_LEFTBAT_POS:
	{
	    if( sscanf( pszBuffer, "%c y=%d", &type, &y ) == 2 )
	    {
		// On met la bonne echelle
		y = y + s_yBatLenOther/2;
		if( s_yHauteurOther ) y = (y * PongGetHauteur()) / s_yHauteurOther;
		y = y - PongGetBatLen() / 2;
		
		// Positionnement de la raquette gauche
		PongSetLeftBatPos( y );
	    }
	    break;
	}
    case SEND_BALL_POS:
	{
	    if( sscanf( pszBuffer, "%c x=%d y=%d", &type, &x, &y ) == 3 )
	    {
		// On met la bonne echelle
		if( s_xLargeurOther ) x = (x * PongGetLargeur()) / s_xLargeurOther;
		if( s_yHauteurOther ) y = (y * PongGetHauteur()) / s_yHauteurOther;
		
		// Positionnement de la balle
		PongSetBallPos( x, y );
	    }
	    break;
	}
    case SEND_SERVEUR_STATUS:
	{
	    GameStatus ServerStatus;
	    if( sscanf( pszBuffer, "%c %d", &type, &ServerStatus ) == 2 )
	    {
		// Partie gagnee / perdu
		*pStatus = (ServerStatus == GAGNE ? PERDU : GAGNE);
	    }
	    break;
	}
    default:
	{
	    // Ne devrait jamais arriver
	    break;
	}
    }
    return ret;
}

////////////////////////////////////////////////////////////////
// Reception de donnees : On isole les donnees par ligne et 
// on appelle PongSocketEvtRecvLine() qui se charge de 
// decoder.
// A appele en reponse a l'evenement FD_READ
////////////////////////////////////////////////////////////////
int PongSocketEvtRecv( GameStatus *pStatus )
{
    char BufferRecv[256];
    int cbRecv;
    char BufferTemp[512];
    char *p, *pBuffer;

    // Reception de la donnee
    cbRecv = SocketRecv( s_hSocket, BufferRecv, sizeof(BufferRecv) - 1 );
    
    // Rien a lire (0) ou bien deconnexion (-1)
    if( cbRecv <= 0 ) return cbRecv;

    // Si on a des donnees pendantes, on commence deja par 
    // les traiter
    if( s_cbBufferRecv )
    {
	memcpy( BufferTemp, s_szBufferRecv, s_cbBufferRecv );
	memcpy( BufferTemp + s_cbBufferRecv, BufferRecv, cbRecv );
	cbRecv += s_cbBufferRecv;
	BufferTemp[cbRecv] = 0;
	s_cbBufferRecv = 0;
    }
    else
    {
	memcpy( BufferTemp, BufferRecv, cbRecv );
	BufferTemp[cbRecv] = 0;
    }

    // Reception data : On traite les donnees ligne par ligne
    // cad qu'on recherche le \n
    pBuffer = BufferTemp;
    *pStatus = ENCOURS;
    while( (*pStatus == ENCOURS) &&
	   (p = strstr( pBuffer, "\n" )) != NULL )
    {
	*p = 0;

	PongSocketEvtRecvLine( pBuffer, pStatus );
	pBuffer = p + 1;
    }

    // S'il reste des lignes incompletes, on les sauvegarde
    s_cbBufferRecv = strlen(pBuffer);
    memcpy( s_szBufferRecv, pBuffer, s_cbBufferRecv );

    return 1;
}

